﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Contracts;
using System.Linq;
using System.Net.NetworkInformation;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Algorithms.IteratedSymbolicExpressionConstruction {
  // this class is a base class for symbolic expression construction policies
  // it filters allowed actions based on tree length constraint
  // and keeps all visited derivation paths in memory to prevent sampling of duplicates
  [StorableClass]
  public abstract class SymbolicExpressionConstructionPolicyBase : ParameterizedNamedItem, ISymbolicExpressionConstructionPolicy {
    [Storable]
    public SymbolicExpressionTreeProblem Problem { get; private set; }
    [Storable]
    public IRandom Random { get; private set; }
    private SearchTree searchTree;

    private class Slot {
      public ISymbolicExpressionTreeNode parent;
      public int childIdx;
      public int minSize;
    }

    [StorableHook(HookType.AfterDeserialization)]
    private void AfterDeserialization() {
      searchTree = new SearchTree();
    }
    protected SymbolicExpressionConstructionPolicyBase(SymbolicExpressionConstructionPolicyBase original, Cloner cloner)
      : base(original, cloner) {
      this.Problem = cloner.Clone(original.Problem);
      this.Random = cloner.Clone(original.Random);

      // search tree is not cloned or stored
      searchTree = new SearchTree();
    }

    [StorableConstructor]
    protected SymbolicExpressionConstructionPolicyBase(bool deserializing) : base(deserializing) { }

    protected SymbolicExpressionConstructionPolicyBase() {
    }

    public virtual void Initialize(SymbolicExpressionTreeProblem problem, IRandom random) {
      this.Problem = problem;
      this.Random = random;
      this.searchTree = new SearchTree(); // represents all realized actionSequences as a prefix tree 
    }

    public ISymbolicExpressionTree Sample(out IEnumerable<Tuple<object, int>> stateActionSequence) {
      var actions = new List<int>();
      var states = new List<object>();

      var openSlots = new Stack<Slot>();
      var g = Problem.Encoding.Grammar;
      var root = g.ProgramRootSymbol.CreateTreeNode();
      int maxDepth = Problem.Encoding.TreeDepthParameter.Value.Value;
      int maxLen = Problem.Encoding.TreeLengthParameter.Value.Value;


      Contract.Assert(Problem.Encoding.FunctionDefinitions == 0);
      openSlots.Push(new Slot() { parent = root, childIdx = 0, minSize = 2 }); // at least two nodes are necessary below root

      // tree size lower bound is the current tree size + the sum of the minimal size for all open slots
      int treeSize = 1;

      searchTree.ResetPosition();

      while (openSlots.Any()) {
        var next = openSlots.Pop();
        var parent = next.parent;
        var childIdx = next.childIdx;

        // states might be defined differently be different policies
        // this allows policies to calculate the state as a function of the current tree and the position where it is changed,
        // or as a function of the list of actions so far, 
        // or as a function of both
        var currentState = CreateState(root, actions, parent, childIdx);
        states.Add(currentState);


        // TODO: only filter the first time later use info from search tree
        var alts = g.GetAllowedChildSymbols(parent.Symbol, childIdx)
          .Where(a => treeSize + g.GetMinimumExpressionLength(a) + openSlots.Select(e => e.minSize).Sum() <= maxLen)
          .ToArray();

        if (searchTree.IsLeafNode()) {
          searchTree.ExpandCurrentNode(alts);
        }

        if (!searchTree.PossibleActions.Any()) {
          throw new InvalidProgramException(string.Format("Couldn't construct a valid tree of maximum length {0} or all possible trees have been visited", maxLen));
        }

        // select a symbol randomly for the child 
        // select random alternative
        var selectedIdx = Select(currentState, searchTree.PossibleActions, Random);
        actions.Add(selectedIdx);

        // and add child node to parent
        var childNode = alts[selectedIdx].CreateTreeNode();
        if (childNode.HasLocalParameters) childNode.ResetLocalParameters(Random);

        Contract.Assert(parent.SubtreeCount == childIdx);
        parent.AddSubtree(childNode); // enforce left-canonical derivation
        treeSize++;

        // push new slots 
        for (int chIdx = childNode.Symbol.MinimumArity - 1; chIdx >= 0; chIdx--) {
          int minForChild = g.GetAllowedChildSymbols(childNode.Symbol, chIdx).Min(a => g.GetMinimumExpressionLength(a)); // min length of all possible alts for the slot
          openSlots.Push(new Slot() { parent = childNode, childIdx = chIdx, minSize = minForChild });
        }

        // if this is the last slot we never have to revisit selectedIdx
        if (!openSlots.Any()) {
          searchTree.RemoveBranch(selectedIdx);
        } else {
          searchTree.Follow(selectedIdx);
        }
      }

      stateActionSequence = states.Zip(actions, Tuple.Create);
      return new SymbolicExpressionTree(root);
    }


    protected abstract int Select(object state, IEnumerable<int> possibleActions, IRandom random);
    public abstract void Update(IEnumerable<Tuple<object, int>> stateActionSequence, double quality);

    protected abstract object CreateState(ISymbolicExpressionTreeNode root, List<int> actions, ISymbolicExpressionTreeNode parent, int childIdx);
  }
}
